package cz.muni.fi.pa165.rest;

import cz.muni.fi.pa165.data.api.rating.RatingCreateDto;
import cz.muni.fi.pa165.data.api.rating.RatingViewDto;
import cz.muni.fi.pa165.data.exception.EntityNotFoundException;
import cz.muni.fi.pa165.data.exception.EntityValidationException;
import cz.muni.fi.pa165.data.model.Rating;
import cz.muni.fi.pa165.data.repository.CriterionRepository;
import cz.muni.fi.pa165.data.repository.MovieRepository;
import cz.muni.fi.pa165.data.repository.RatingRepository;
import cz.muni.fi.pa165.data.repository.UserRepository;
import cz.muni.fi.pa165.service.RatingService;
import io.swagger.v3.oas.annotations.Operation;
import io.swagger.v3.oas.annotations.Parameter;
import io.swagger.v3.oas.annotations.responses.ApiResponse;
import org.modelmapper.ModelMapper;
import org.springdoc.core.annotations.ParameterObject;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.Pageable;
import org.springframework.http.HttpStatus;
import org.springframework.http.MediaType;
import org.springframework.http.ResponseEntity;
import org.springframework.web.bind.annotation.*;

import java.util.List;
import java.util.Optional;

@RestController
@RequestMapping(path = "/rating", produces = MediaType.APPLICATION_JSON_VALUE)
public class RatingController {

    private final RatingRepository ratingRepository;
    private final RatingService ratingService;
    private final ModelMapper modelMapper;
    private final UserRepository userRepository;
    private final MovieRepository movieRepository;
    private final CriterionRepository criterionRepository;

    @Autowired
    public RatingController(RatingRepository ratingRepository, RatingService ratingService, ModelMapper modelMapper, UserRepository userRepository, MovieRepository movieRepository, CriterionRepository criterionRepository) {
        this.ratingRepository = ratingRepository;
        this.ratingService = ratingService;
        this.modelMapper = modelMapper;
        this.userRepository = userRepository;
        this.movieRepository = movieRepository;
        this.criterionRepository = criterionRepository;
    }

    @GetMapping
    @ResponseStatus(HttpStatus.OK)
    @Operation(
            summary = "Find ratings",
            description = "Gets all ratings according to the specified query parameters.",
            parameters = {
                    @Parameter(name = "userId", description = "User ID", example = "1"),
                    @Parameter(name = "movieId", description = "Movie ID", example = "1"),
                    @Parameter(name = "criterionId", description = "Criterion ID", example = "1")
            },
            responses = {
                    @ApiResponse(responseCode = "200", description = "Ratings found")
            },
            tags = {"Rating"}
    )
    public ResponseEntity<List<RatingViewDto>> getAll(@RequestParam Optional<Integer> userId, @RequestParam Optional<Integer> movieId, @RequestParam Optional<Integer> criterionId, @ParameterObject Pageable pageable) {
        Page<Rating> ratings = ratingRepository.findAllByUserIdAndMovieIdAndCriterionId(userId, movieId, criterionId, pageable);
        List<RatingViewDto> view = ratings.getContent().stream().map(rating -> modelMapper.map(rating, RatingViewDto.class)).toList();
        return ResponseEntity.ok(view);
    }

    @PutMapping
    @ResponseStatus(HttpStatus.OK)
    @Operation(
            summary = "Create or update a rating",
            description = "Create a new rating with specified value or edit existing one.",
            responses = {
                    @ApiResponse(responseCode = "200", description = "Rating created or updated"),
                    @ApiResponse(responseCode = "400", description = "Rating value is not between 0 and 10, user, movie or criterion does not exist")
            },
            tags = {"Rating"}
    )
    public ResponseEntity<RatingViewDto> add(@RequestBody RatingCreateDto rating) {
        Rating entity = modelMapper.map(rating, Rating.class);

        if (entity.getValue() < 0 || entity.getValue() > 10) {
            throw new EntityValidationException("Rating must be between 0 and 10.");
        }

        if (userRepository.findById(entity.getUser().getId()).isEmpty()) {
            throw new EntityNotFoundException("User", entity.getUser().getId());
        }

        if (movieRepository.findById(entity.getMovie().getId()).isEmpty()) {
            throw new EntityNotFoundException("Movie", entity.getMovie().getId());
        }

        if (criterionRepository.findById(entity.getCriterion().getId()).isEmpty()) {
            throw new EntityNotFoundException("Criterion", entity.getCriterion().getId());
        }

        Optional<Rating> exists = ratingRepository.findByUserIdAndMovieIdAndCriterionId(rating.getUserId(), rating.getMovieId(), rating.getCriterionId());

        if (exists.isPresent()) {
            entity.setId(exists.get().getId());
        }

        entity = ratingRepository.save(entity);
        RatingViewDto view = modelMapper.map(entity, RatingViewDto.class);

        return ResponseEntity.ok(view);
    }

    @DeleteMapping
    @ResponseStatus(HttpStatus.OK)
    @Operation(
            summary = "Delete rating",
            description = "Delete a specific rating.",
            parameters = {
                    @Parameter(name = "userId", description = "User ID", example = "1"),
                    @Parameter(name = "movieId", description = "Movie ID", example = "1"),
                    @Parameter(name = "criterionId", description = "Criterion ID", example = "1")
            },
            responses = {
                    @ApiResponse(responseCode = "200", description = "Rating deleted"),
                    @ApiResponse(responseCode = "404", description = "Rating not found")
            },
            tags = {"Rating"}
    )
    public ResponseEntity<RatingViewDto> delete(@RequestParam int userId, @RequestParam int movieId, @RequestParam int criterionId) {
        Rating entity = ratingRepository.findByUserIdAndMovieIdAndCriterionId(userId, movieId, criterionId)
                .orElseThrow(() -> new EntityNotFoundException("Matching rating was not found."));

        ratingRepository.delete(entity);

        RatingViewDto view = modelMapper.map(entity, RatingViewDto.class);
        return ResponseEntity.ok(view);
    }

}
